"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.runScript = exports.handler = exports.help = exports.commandNames = exports.completion = exports.cliOptionsTypes = exports.rcOptionsTypes = exports.shorthands = exports.REPORT_SUMMARY_OPTION_HELP = exports.SEQUENTIAL_OPTION_HELP = exports.RESUME_FROM_OPTION_HELP = exports.PARALLEL_OPTION_HELP = exports.IF_PRESENT_OPTION_HELP = exports.IF_PRESENT_OPTION = void 0;
const path_1 = __importDefault(require("path"));
const p_limit_1 = __importDefault(require("p-limit"));
const cli_utils_1 = require("@pnpm/cli-utils");
const common_cli_options_help_1 = require("@pnpm/common-cli-options-help");
const config_1 = require("@pnpm/config");
const error_1 = require("@pnpm/error");
const lifecycle_1 = require("@pnpm/lifecycle");
const pick_1 = __importDefault(require("ramda/src/pick"));
const realpath_missing_1 = __importDefault(require("realpath-missing"));
const render_help_1 = __importDefault(require("render-help"));
const runRecursive_1 = require("./runRecursive");
const existsInDir_1 = require("./existsInDir");
const exec_1 = require("./exec");
exports.IF_PRESENT_OPTION = {
    'if-present': Boolean,
};
exports.IF_PRESENT_OPTION_HELP = {
    description: 'Avoid exiting with a non-zero exit code when the script is undefined',
    name: '--if-present',
};
exports.PARALLEL_OPTION_HELP = {
    description: 'Completely disregard concurrency and topological sorting, \
running a given script immediately in all matching packages \
with prefixed streaming output. This is the preferred flag \
for long-running processes such as watch run over many packages.',
    name: '--parallel',
};
exports.RESUME_FROM_OPTION_HELP = {
    description: 'Command executed from given package',
    name: '--resume-from',
};
exports.SEQUENTIAL_OPTION_HELP = {
    description: 'Run the specified scripts one by one',
    name: '--sequential',
};
exports.REPORT_SUMMARY_OPTION_HELP = {
    description: 'Save the execution results of every package to "pnpm-exec-summary.json". Useful to inspect the execution time and status of each package.',
    name: '--report-summary',
};
exports.shorthands = {
    parallel: [
        '--workspace-concurrency=Infinity',
        '--no-sort',
        '--stream',
        '--recursive',
    ],
    sequential: [
        '--workspace-concurrency=1',
    ],
};
function rcOptionsTypes() {
    return {
        ...(0, pick_1.default)([
            'npm-path',
            'use-node-version',
        ], config_1.types),
    };
}
exports.rcOptionsTypes = rcOptionsTypes;
function cliOptionsTypes() {
    return {
        ...(0, pick_1.default)([
            'bail',
            'sort',
            'unsafe-perm',
            'use-node-version',
            'workspace-concurrency',
            'scripts-prepend-node-path',
        ], config_1.types),
        ...exports.IF_PRESENT_OPTION,
        recursive: Boolean,
        reverse: Boolean,
        'resume-from': String,
        'report-summary': Boolean,
    };
}
exports.cliOptionsTypes = cliOptionsTypes;
const completion = async (cliOpts, params) => {
    if (params.length > 0) {
        return [];
    }
    const manifest = await (0, cli_utils_1.readProjectManifestOnly)(cliOpts.dir ?? process.cwd(), cliOpts);
    return Object.keys(manifest.scripts ?? {}).map((name) => ({ name }));
};
exports.completion = completion;
exports.commandNames = ['run', 'run-script'];
function help() {
    return (0, render_help_1.default)({
        aliases: ['run-script'],
        description: 'Runs a defined package script.',
        descriptionLists: [
            {
                title: 'Options',
                list: [
                    {
                        description: 'Run the defined package script in every package found in subdirectories \
or every workspace package, when executed inside a workspace. \
For options that may be used with `-r`, see "pnpm help recursive"',
                        name: '--recursive',
                        shortAlias: '-r',
                    },
                    {
                        description: 'The command will exit with a 0 exit code even if the script fails',
                        name: '--no-bail',
                    },
                    exports.IF_PRESENT_OPTION_HELP,
                    exports.PARALLEL_OPTION_HELP,
                    exports.RESUME_FROM_OPTION_HELP,
                    ...common_cli_options_help_1.UNIVERSAL_OPTIONS,
                    exports.SEQUENTIAL_OPTION_HELP,
                    exports.REPORT_SUMMARY_OPTION_HELP,
                ],
            },
            common_cli_options_help_1.FILTERING,
        ],
        url: (0, cli_utils_1.docsUrl)('run'),
        usages: ['pnpm run <command> [<args>...]'],
    });
}
exports.help = help;
async function handler(opts, params) {
    let dir;
    const [scriptName, ...passedThruArgs] = params;
    if (opts.recursive) {
        if (scriptName || Object.keys(opts.selectedProjectsGraph).length > 1) {
            return (0, runRecursive_1.runRecursive)(params, opts);
        }
        dir = Object.keys(opts.selectedProjectsGraph)[0];
    }
    else {
        dir = opts.dir;
    }
    const manifest = await (0, cli_utils_1.readProjectManifestOnly)(dir, opts);
    if (!scriptName) {
        const rootManifest = opts.workspaceDir && opts.workspaceDir !== dir
            ? (await (0, cli_utils_1.tryReadProjectManifest)(opts.workspaceDir, opts)).manifest
            : undefined;
        return printProjectCommands(manifest, rootManifest ?? undefined);
    }
    const specifiedScripts = getSpecifiedScripts(manifest.scripts ?? {}, scriptName);
    if (specifiedScripts.length < 1) {
        if (opts.ifPresent)
            return;
        if (opts.fallbackCommandUsed) {
            if (opts.argv == null)
                throw new Error('Could not fallback because opts.argv.original was not passed to the script runner');
            return (0, exec_1.handler)({
                selectedProjectsGraph: {},
                ...opts,
            }, opts.argv.original.slice(1));
        }
        if (opts.workspaceDir) {
            const { manifest: rootManifest } = await (0, cli_utils_1.tryReadProjectManifest)(opts.workspaceDir, opts);
            if (getSpecifiedScripts(rootManifest?.scripts ?? {}, scriptName).length > 0 && specifiedScripts.length < 1) {
                throw new error_1.PnpmError('NO_SCRIPT', `Missing script: ${scriptName}`, {
                    hint: `But script matched with ${scriptName} is present in the root of the workspace,
so you may run "pnpm -w run ${scriptName}"`,
                });
            }
        }
        throw new error_1.PnpmError('NO_SCRIPT', `Missing script: ${scriptName}`);
    }
    const lifecycleOpts = {
        depPath: dir,
        extraBinPaths: opts.extraBinPaths,
        extraEnv: opts.extraEnv,
        pkgRoot: dir,
        rawConfig: opts.rawConfig,
        rootModulesDir: await (0, realpath_missing_1.default)(path_1.default.join(dir, 'node_modules')),
        scriptsPrependNodePath: opts.scriptsPrependNodePath,
        scriptShell: opts.scriptShell,
        silent: opts.reporter === 'silent',
        shellEmulator: opts.shellEmulator,
        stdio: 'inherit',
        unsafePerm: true, // when running scripts explicitly, assume that they're trusted.
    };
    const existsPnp = existsInDir_1.existsInDir.bind(null, '.pnp.cjs');
    const pnpPath = (opts.workspaceDir && await existsPnp(opts.workspaceDir)) ??
        await existsPnp(dir);
    if (pnpPath) {
        lifecycleOpts.extraEnv = {
            ...lifecycleOpts.extraEnv,
            ...(0, lifecycle_1.makeNodeRequireOption)(pnpPath),
        };
    }
    try {
        const limitRun = (0, p_limit_1.default)(opts.workspaceConcurrency ?? 4);
        const _runScript = exports.runScript.bind(null, { manifest, lifecycleOpts, runScriptOptions: { enablePrePostScripts: opts.enablePrePostScripts ?? false }, passedThruArgs });
        await Promise.all(specifiedScripts.map(script => limitRun(() => _runScript(script))));
    }
    catch (err) { // eslint-disable-line
        if (opts.bail !== false) {
            throw err;
        }
    }
    return undefined;
}
exports.handler = handler;
const ALL_LIFECYCLE_SCRIPTS = new Set([
    'prepublish',
    'prepare',
    'prepublishOnly',
    'prepack',
    'postpack',
    'publish',
    'postpublish',
    'preinstall',
    'install',
    'postinstall',
    'preuninstall',
    'uninstall',
    'postuninstall',
    'preversion',
    'version',
    'postversion',
    'pretest',
    'test',
    'posttest',
    'prestop',
    'stop',
    'poststop',
    'prestart',
    'start',
    'poststart',
    'prerestart',
    'restart',
    'postrestart',
    'preshrinkwrap',
    'shrinkwrap',
    'postshrinkwrap',
]);
function printProjectCommands(manifest, rootManifest) {
    const lifecycleScripts = [];
    const otherScripts = [];
    for (const [scriptName, script] of Object.entries(manifest.scripts ?? {})) {
        if (ALL_LIFECYCLE_SCRIPTS.has(scriptName)) {
            lifecycleScripts.push([scriptName, script]);
        }
        else {
            otherScripts.push([scriptName, script]);
        }
    }
    if (lifecycleScripts.length === 0 && otherScripts.length === 0) {
        return 'There are no scripts specified.';
    }
    let output = '';
    if (lifecycleScripts.length > 0) {
        output += `Lifecycle scripts:\n${renderCommands(lifecycleScripts)}`;
    }
    if (otherScripts.length > 0) {
        if (output !== '')
            output += '\n\n';
        output += `Commands available via "pnpm run":\n${renderCommands(otherScripts)}`;
    }
    if ((rootManifest?.scripts) == null) {
        return output;
    }
    const rootScripts = Object.entries(rootManifest.scripts);
    if (rootScripts.length === 0) {
        return output;
    }
    if (output !== '')
        output += '\n\n';
    output += `Commands of the root workspace project (to run them, use "pnpm -w run"):
${renderCommands(rootScripts)}`;
    return output;
}
const runScript = async function (opts, scriptName) {
    if (opts.runScriptOptions.enablePrePostScripts &&
        opts.manifest.scripts?.[`pre${scriptName}`] &&
        !opts.manifest.scripts[scriptName].includes(`pre${scriptName}`)) {
        await (0, lifecycle_1.runLifecycleHook)(`pre${scriptName}`, opts.manifest, opts.lifecycleOpts);
    }
    await (0, lifecycle_1.runLifecycleHook)(scriptName, opts.manifest, { ...opts.lifecycleOpts, args: opts.passedThruArgs });
    if (opts.runScriptOptions.enablePrePostScripts &&
        opts.manifest.scripts?.[`post${scriptName}`] &&
        !opts.manifest.scripts[scriptName].includes(`post${scriptName}`)) {
        await (0, lifecycle_1.runLifecycleHook)(`post${scriptName}`, opts.manifest, opts.lifecycleOpts);
    }
};
exports.runScript = runScript;
function renderCommands(commands) {
    return commands.map(([scriptName, script]) => `  ${scriptName}\n    ${script}`).join('\n');
}
function getSpecifiedScripts(scripts, scriptName) {
    const specifiedSelector = (0, runRecursive_1.getSpecifiedScripts)(scripts, scriptName);
    if (specifiedSelector.length > 0) {
        return specifiedSelector;
    }
    // if a user passes start command as scriptName, `node server.js` will be executed as a fallback, so return start command even if start command is not defined in package.json
    if (scriptName === 'start') {
        return [scriptName];
    }
    return [];
}
//# sourceMappingURL=run.js.map